@@ -47,6 +47,20 @@ impl Global {
     }
 }
 
+/// Marks a [domain family](Domain) that uniquely characterizes a [domain instance](Domain).
+///
+/// See [`Global`] and [`unique_domain`] for examples of families that safely manage this.
+///
+/// # Safety
+///
+/// Implementors of this trait must guarantee only one Domain of the implementing family can ever be constructed.
+pub unsafe trait Singleton {}
+
+// Safety: we can guarantee that there's only ever one Domain<Global> because Global itself is not
+// possible to construct outside of this crate (due to #[non_exhaustive] + no public constructor),
+// and we only ever construct one Domain from it internally in the form of a single static.
+unsafe impl Singleton for Global {}
+
 #[cfg(not(loom))]
 static SHARED_DOMAIN: Domain<Global> = Domain::new(&Global::new());
 
@@ -133,19 +147,34 @@ pub struct Domain<F> {
     shutdown: bool,
 }
 
+#[cfg(miri)]
+extern "Rust" {
+    fn miri_static_root(ptr: *const u8);
+}
+
 impl Domain<Global> {
     /// Get a handle to the singleton [global domain](Global).
     pub fn global() -> &'static Self {
+        #[cfg(miri)]
+        unsafe {
+            miri_static_root(&SHARED_DOMAIN as *const _ as *const u8);
+        };
+
         &SHARED_DOMAIN
     }
 }
 
 /// Generate a [`Domain`] with an entirely unique domain family.
+///
+/// The generated family implements [`Singleton`], which enables the use of [`AtomicPtr::safe_load`].
 #[macro_export]
 macro_rules! unique_domain {
-    () => {
-        Domain::new(&|| {})
-    };
+    () => {{
+        struct UniqueFamily;
+        // Safety: nowhere else can construct an instance of UniqueFamily to pass to Domain::new.
+        unsafe impl Singleton for UniqueFamily {}
+        Domain::new(&UniqueFamily)
+    }};
 }
 
 // Macro to make new const only when not in loom.
@@ -176,7 +205,7 @@ macro_rules! new {
             Self {
                 hazptrs: HazPtrRecords {
                     head: AtomicPtr::new(core::ptr::null_mut()),
-                    head_available: AtomicUsize::new(0),
+                    head_available: AtomicPtr::new(core::ptr::null_mut()),
                     count: AtomicIsize::new(0),
                 },
                 untagged,
@@ -247,24 +276,28 @@ impl<F> Domain<F> {
 
         loop {
             let avail = self.hazptrs.head_available.load(Ordering::Acquire);
-            if avail == core::ptr::null::<HazPtrRecord>() as usize {
-                return (core::ptr::null_mut(), 0);
+            if avail.is_null() {
+                return (avail, 0);
             }
-            debug_assert_ne!(avail, core::ptr::null::<HazPtrRecord>() as usize | LOCK_BIT);
+            debug_assert_ne!(avail, LOCK_BIT as *mut _);
             if (avail as usize & LOCK_BIT) == 0 {
-                // Definitely a valid pointer now.
-                let avail: *const HazPtrRecord = avail as _;
-
                 // The available list is not currently locked.
                 //
                 // XXX: This could be a fetch_or and allow progress even if there's a new (but
-                // unlocked) head.
+                // unlocked) head. However, `AtomicPtr` doesn't support fetch_or at the moment, so
+                // we'd have to convert it to an `AtomicUsize`. This will in turn make Miri fail
+                // (with -Zmiri-tag-raw-pointers, which we want enabled) to track the provenance of
+                // the pointer in question through the int-to-ptr conversion. The workaround is
+                // probably to mock a type that is `AtomicUsize` with `fetch_or` with
+                // `#[cfg(not(miri))]`, but is `AtomicPtr` with `compare_exchange` with
+                // `#[cfg(miri)]`. It ain't pretty, but should do the job. The issue is tracked in
+                // https://github.com/rust-lang/miri/issues/1993.
                 if self
                     .hazptrs
                     .head_available
                     .compare_exchange_weak(
-                        avail as usize,
-                        avail as usize | LOCK_BIT,
+                        avail,
+                        with_lock_bit(avail),
                         Ordering::AcqRel,
                         Ordering::Relaxed,
                     )
@@ -307,9 +340,7 @@ impl<F> Domain<F> {
         }
 
         // NOTE: This releases the lock
-        self.hazptrs
-            .head_available
-            .store(next as usize, Ordering::Release);
+        self.hazptrs.head_available.store(next, Ordering::Release);
         unsafe { &*tail }
             .available_next
             .store(core::ptr::null_mut(), Ordering::Relaxed);
@@ -325,7 +356,7 @@ impl<F> Domain<F> {
         debug_assert_eq!(head as *const _ as usize & LOCK_BIT, 0);
         loop {
             let avail = self.hazptrs.head_available.load(Ordering::Acquire);
-            if (avail & LOCK_BIT) == 0 {
+            if (avail as usize & LOCK_BIT) == 0 {
                 tail.available_next
                     .store(avail as *mut _, Ordering::Relaxed);
                 if self
@@ -333,7 +364,7 @@ impl<F> Domain<F> {
                     .head_available
                     .compare_exchange_weak(
                         avail,
-                        head as *const _ as usize,
+                        head as *const _ as *mut _,
                         Ordering::AcqRel,
                         Ordering::Relaxed,
                     )
@@ -386,6 +417,8 @@ impl<F> Domain<F> {
 
     /// Retire `ptr`, and reclaim it once it is safe to do so.
     ///
+    /// `T` must be `Send` since it may be reclaimed by a different thread.
+    ///
     /// # Safety
     ///
     /// 1. no [`HazardPointer`] will guard `ptr` from this point forward.
@@ -393,6 +426,7 @@ impl<F> Domain<F> {
     /// 3. `ptr` is valid as `&T` until `self` is dropped.
     pub unsafe fn retire_ptr<T, P>(&self, ptr: *mut T) -> usize
     where
+        T: Send,
         P: Pointer<T>,
     {
         // First, stick ptr onto the list of retired objects.
@@ -631,12 +665,12 @@ impl<F> Domain<F> {
         }
     }
 
-    #[cfg(loom)]
+    #[cfg(any(loom, miri))]
     fn now() -> u64 {
         0
     }
 
-    #[cfg(all(feature = "std", target_pointer_width = "64", not(loom)))]
+    #[cfg(all(feature = "std", target_pointer_width = "64", not(loom), not(miri)))]
     fn now() -> u64 {
         u64::try_from(
             std::time::SystemTime::now()
@@ -714,7 +748,7 @@ impl<F> Drop for Domain<F> {
 
 struct HazPtrRecords {
     head: AtomicPtr<HazPtrRecord>,
-    head_available: AtomicUsize, // really *mut HazPtrRecord
+    head_available: AtomicPtr<HazPtrRecord>,
     count: AtomicIsize,
 }
 
@@ -805,20 +839,18 @@ impl RetiredList {
     }
 }
 
-/*
-fn foo() {
-    let domain = Domain::new();
-    'a: {
-        let d: &'a Domain = &domain;
-        let t = String::new();
-        let z: Box<PrintOnDrop<&'a String>> = Box::new(PrintOnDrop(&t));
-        d.retire(z); // z goes on .retired, but is _not_ dropped
-        // drop(t), so z is no longer valid
-    }
-    // walk .retired, find that z can be reclaimed, call drop(z);
-    drop(domain);
+// Helpers to set and unset the lock bit on a `*mut HazPtrRecord` without losing pointer
+// provenance. See https://github.com/rust-lang/miri/issues/1993 for details.
+fn with_lock_bit(ptr: *mut HazPtrRecord) -> *mut HazPtrRecord {
+    int_to_ptr_with_provenance(ptr as usize | LOCK_BIT, ptr)
+}
+fn without_lock_bit(ptr: *mut HazPtrRecord) -> *mut HazPtrRecord {
+    int_to_ptr_with_provenance(ptr as usize & !LOCK_BIT, ptr)
+}
+fn int_to_ptr_with_provenance<T>(addr: usize, prov: *mut T) -> *mut T {
+    let ptr = prov.cast::<u8>();
+    ptr.wrapping_add(addr.wrapping_sub(ptr as usize)).cast()
 }
-*/
 
 /// ```compile_fail
 /// use haphazard::*;
@@ -869,7 +901,7 @@ struct CannotConfuseGlobalReader;
 /// let mut h = HazardPointer::new_in_domain(&dr);
 ///
 /// // This shouldn't compile because families differ.
-/// let _ = unsafe { x.load(&mut h).expect("not null") };
+/// let _ = x.safe_load(&mut h).expect("not null");
 /// ```
 #[cfg(doctest)]
 struct CannotConfuseAcrossFamilies;
