@@ -105,6 +105,10 @@
 //! drop(h);
 //! let n = Domain::global().eager_reclaim();
 //! assert_eq!(n, 1);
+//!
+//! // Remember to also retire the item stored in the AtomicPtr when it's dropped
+//! // (assuming of course that the pointer is not shared elsewhere):
+//! unsafe { x.retire(); }
 //! ```
 //!
 //! # Differences from the specification
@@ -175,6 +179,7 @@ use core::sync::atomic::Ordering;
 
 pub use domain::Domain;
 pub use domain::Global;
+pub use domain::Singleton;
 pub use hazard::{HazardPointer, HazardPointerArray};
 
 /// A managed pointer type which can be safely shared between threads.
@@ -189,7 +194,11 @@ pub use hazard::{HazardPointer, HazardPointerArray};
 ///
 /// ```rust
 /// # use haphazard::AtomicPtr;
-/// let _: AtomicPtr<usize> = AtomicPtr::from(Box::new(42));
+/// let ptr: AtomicPtr<usize> = AtomicPtr::from(Box::new(42));
+///
+/// // Remember to retire the item stored in the AtomicPtr when it's dropped
+/// // (assuming of course that the pointer is not also shared elsewhere):
+/// unsafe { ptr.retire(); }
 /// ```
 ///
 /// Note the explicit use of `AtomicPtr<usize>`, which is needed to get the default values for the
@@ -202,6 +211,10 @@ pub use hazard::{HazardPointer, HazardPointerArray};
 /// This type has the same in-memory representation as a
 /// [`std::sync::AtomicPtr`](core::sync::AtomicPtr).
 ///
+/// **Warning:** When this type is dropped, it does _not_ automatically retire the object it is
+/// currently pointing to. In order to retire (and eventually reclaim) that object, use
+/// [`AtomicPtr::retire_in`].
+///
 /// **Note:** This type is only available on platforms that support atomic loads and stores of
 /// pointers. Its size depends on the target pointer’s size.
 ///
@@ -224,6 +237,21 @@ pub struct AtomicPtr<T, F = domain::Global, P = alloc::boxed::Box<T>>(
     PhantomData<(F, *mut P)>,
 );
 
+// # Safety
+//
+// It's safe to give away ownership of an AtomicPtr to a different thread, because it holds no
+// state that would be invalidated by giving ownership to another thread. Basically, this type is
+// Sync because std::sync::atomic::AtomicPtr is Sync.
+unsafe impl<T, F, P> Send for AtomicPtr<T, F, P> {}
+
+// # Safety
+//
+// It's safe to share an AtomicPtr between threads, because it internally ensures that such
+// accesses are correctly coordinated by using the atomic instructions of AtomicPtr. Furthermore,
+// all the accessors of `AtomicPtr` that yield `&T` ensure that `T: Sync`, and any that retire `T`
+// (and thus may take ownership of one) ensure that `T: Send`.
+unsafe impl<T, F, P> Sync for AtomicPtr<T, F, P> {}
+
 impl<T, F, P> From<P> for AtomicPtr<T, F, P>
 where
     P: raw::Pointer<T>,
@@ -325,11 +353,16 @@ where
 {
     /// Retire the referenced object, and reclaim it once it is safe to do so.
     ///
+    /// `T` must be `Send` since it may be reclaimed by a different thread.
+    ///
     /// # Safety
     ///
     /// 1. The pointed-to object will never again be returned by any [`AtomicPtr::load`].
     /// 2. The pointed-to object has not already been retired.
-    pub unsafe fn retire(self) -> usize {
+    pub unsafe fn retire(self) -> usize
+    where
+        T: Send,
+    {
         // Safety:
         //
         // 1. Same as our caller requirement #1.
@@ -348,6 +381,8 @@ where
     /// Retire the referenced object, and reclaim it once it is safe to do so, through the given
     /// `domain`.
     ///
+    /// `T` must be `Send` since it may be reclaimed by a different thread.
+    ///
     /// # Safety
     ///
     /// 1. The pointed-to object will never again be returned by any [`AtomicPtr::load`].
@@ -358,7 +393,10 @@ where
     /// Note that requirement #3 is _partially_ enforced by the domain family (`F`), but it's on
     /// you to ensure that you don't "cross the streams" between multiple `Domain<F>`, if those can
     /// arise in your application.
-    pub unsafe fn retire_in(self, domain: &Domain<F>) -> usize {
+    pub unsafe fn retire_in(self, domain: &Domain<F>) -> usize
+    where
+        T: Send,
+    {
         // Safety:
         //
         // 1. implied by our #1 and #3: if load won't return it, there's no other way to guard it
@@ -369,25 +407,28 @@ where
     }
 }
 
-impl<T, P> AtomicPtr<T, domain::Global, P> {
+impl<T, F, P> AtomicPtr<T, F, P>
+where
+    F: Singleton,
+{
     /// Loads the value from the stored pointer and guards it using the given hazard pointer.
     ///
     /// The guard ensures that the loaded `T` will remain valid for as long as you hold a reference
     /// to it.
     ///
-    /// Note that this method is _only_ available when using [`Domain<Global>`](domain::Global),
-    /// since it is a singleton, and thus is guaranteed to fulfill the safety requirement of
-    /// [`AtomicPtr::load`]. For all other domains, use [`AtomicPtr::load`].
-    pub fn safe_load<'hp>(
-        &'_ self,
-        hp: &'hp mut HazardPointer<'static, domain::Global>,
-    ) -> Option<&'hp T>
+    /// This method is only available for domains with _singleton families_, because
+    /// implementations of the unsafe [`Domain::Singleton`] trait guarantee that there
+    /// is only ever one instance of a Domain with the family in question, which in turn
+    /// implies that loads and stores using such a family **must** be using the same
+    /// (single) instance of that domain.
+    pub fn safe_load<'hp, 'd>(&'_ self, hp: &'hp mut HazardPointer<'d, F>) -> Option<&'hp T>
     where
-        T: 'hp,
+        T: Sync + 'hp,
+        F: 'static,
     {
-        // Safety: since there is exactly one Domain<Global>, we know that all calls to `load` that
-        // have returned this object must have been using the same (global) domain as we're
-        // retiring to.
+        // Safety: by the safety guarantees of Domain::Singleton there is exactly one domain of
+        // this family, so we know that all calls to `load` that have returned this object must
+        // have been using the same domain as we're retiring to.
         unsafe { self.load(hp) }
     }
 }
@@ -408,7 +449,7 @@ impl<T, F, P> AtomicPtr<T, F, P> {
     /// in your application.
     pub unsafe fn load<'hp, 'd>(&'_ self, hp: &'hp mut HazardPointer<'d, F>) -> Option<&'hp T>
     where
-        T: 'hp,
+        T: Sync + 'hp,
         F: 'static,
     {
         unsafe { hp.protect(&self.0) }
@@ -445,11 +486,16 @@ where
 {
     /// Retire the currently-referenced object, and reclaim it once it is safe to do so.
     ///
+    /// `T` must be `Send` since it may be reclaimed by a different thread.
+    ///
     /// # Safety
     ///
     /// 1. The currently-referenced object will never again be returned by any [`AtomicPtr::load`].
     /// 2. The currently-referenced object has not already been retired.
-    pub unsafe fn retire(self) -> usize {
+    pub unsafe fn retire(self) -> usize
+    where
+        T: Send,
+    {
         // Safety:
         //
         // 1. Same as our caller requirement #1.
@@ -468,6 +514,8 @@ where
     /// Retire the currently-referenced object, and reclaim it once it is safe to do so, through
     /// the given `domain`.
     ///
+    /// `T` must be `Send` since it may be reclaimed by a different thread.
+    ///
     /// # Safety
     ///
     /// 1. The currently-referenced object will never again be returned by any [`AtomicPtr::load`].
@@ -478,7 +526,10 @@ where
     /// Note that requirement #3 is _partially_ enforced by the domain family (`F`), but it's on
     /// you to ensure that you don't "cross the streams" between multiple `Domain<F>`, if those can
     /// arise in your application.
-    pub unsafe fn retire_in(self, domain: &Domain<F>) -> usize {
+    pub unsafe fn retire_in(self, domain: &Domain<F>) -> usize
+    where
+        T: Send,
+    {
         let ptr = self.into_inner();
         unsafe { domain.retire_ptr::<T, P>(ptr) }
     }
